#include "HinesMatrix.hpp"
#include "PlatformFunctions.hpp"
#include "HinesStruct.hpp"
#include "Connections.hpp"
#include "SpikeStatistics.hpp"
//#include "cuda_runtime_api.h"
#include <cassert>
#include <cstdlib>
#include <pthread.h>
//#include <cuda.h>

extern __global__ void solveMatrixG(HinesStruct *hList, int nSteps, int nNeurons, ftype *spikeListGlobal, ftype *weightListGlobal, int *spikeListPosGlobal, int *spikeListStartGlobal, ftype *vmListGlobal);

void checkCUDAError(const char *msg)
{
    cudaError_t err = cudaGetLastError();
    if( cudaSuccess != err)
    {
        fprintf(stderr, "Cuda error: %s: %s.\n", msg, cudaGetErrorString( err) );
        exit(EXIT_FAILURE);
    }
}

void prepareSynapses(SharedNeuronGpuData *sharedData, int *nNeurons, int startType, int endType) {

	HinesMatrix **matrixList = sharedData->matrixList;
	HinesStruct **hList = sharedData->hList;
	SynapticData* synData = sharedData->synData;

	/**
	 * Prepare the synaptic channels and spike generation
	 */
	int totalTypes = synData->totalTypes;

	pthread_mutex_lock (sharedData->mutex);
	if (synData->spikeListDevice == 0) {
		synData->spikeListDevice   = (ftype **) malloc (sizeof(ftype *) * totalTypes);
		synData->weightListDevice  = (ftype **) malloc (sizeof(ftype *) * totalTypes);
		synData->spikeListPosDevice  = (int **) malloc (sizeof(int *) * totalTypes);
		synData->spikeListStartDevice = (int **) malloc (sizeof(int *) * totalTypes);


		synData->spikeListGlobal     = (ftype **) malloc (sizeof(ftype *) * totalTypes);
		synData->spikeListPosGlobal  = (int **) malloc (sizeof(int *) * totalTypes);
		synData->spikeListStartGlobal = (int **) malloc (sizeof(int *) * totalTypes);

		synData->spikeTimeListHost    = (ftype **) malloc (sizeof(ftype *) * totalTypes);
		synData->spikeTimeListDevice  = (ftype **) malloc (sizeof(ftype *) * totalTypes);

		synData->vmListHost    = (ftype **) malloc (sizeof(ftype *) * totalTypes);
		synData->vmListDevice  = (ftype **) malloc (sizeof(ftype *) * totalTypes);

		synData->nGeneratedSpikesHost    = (ucomp **) malloc (sizeof(ucomp *) * totalTypes);
		synData->nGeneratedSpikesDevice  = (ucomp **) malloc (sizeof(ucomp *) * totalTypes);
	}
	pthread_mutex_unlock (sharedData->mutex);


	//printf("0\n");

	for (int type = startType; type < endType; type++) {
		synData->spikeListGlobal[type] = 0;
		int totalListPosSize = 1 + matrixList[type][0].synapticChannels->synapseListSize * nNeurons[type];
		synData->spikeListPosGlobal[type]  = (int *) malloc(sizeof(int) * totalListPosSize);
		synData->spikeListStartGlobal[type] = (int *) malloc(sizeof(int) * (nNeurons[type]+1));
		cudaMalloc ((void **) &(synData->spikeListPosDevice[type]),  sizeof(int)  * totalListPosSize);
		cudaMalloc ((void **) &(synData->spikeListStartDevice[type]), sizeof(int)  * (nNeurons[type]+1));

		synData->vmListHost[type] = (ftype *) malloc(sizeof(ftype) * nNeurons[type]);
		cudaMalloc ((void **) &(synData->vmListDevice[type]), sizeof(ftype)  * nNeurons[type]);

		//printf("01\n");

		for (int i=0; i<totalListPosSize; i++)
			synData->spikeListPosGlobal[type][i] = 0;
		for (int i=0; i<nNeurons[type]+1; i++)
			synData->spikeListStartGlobal[type][i] = 0;


		cudaMemcpy(synData->spikeListPosDevice[type], synData->spikeListPosGlobal[type],
				sizeof(int) * totalListPosSize, cudaMemcpyHostToDevice);
		cudaMemcpy(synData->spikeListStartDevice[type], synData->spikeListStartGlobal[type],
				sizeof(int) * (nNeurons[type]+1), cudaMemcpyHostToDevice);
	}

	for (int type = startType; type < endType; type++) {
		int spikeTimeListSize = matrixList[type][0].spikeTimeListSize;

		synData->spikeTimeListHost[type] = (ftype *) malloc(sizeof(ftype) * spikeTimeListSize * nNeurons[type]);
		cudaMalloc ((void **) &(synData->spikeTimeListDevice[type]), sizeof(ftype) * spikeTimeListSize * nNeurons[type]);

		synData->nGeneratedSpikesHost[type] = (ucomp *) malloc(sizeof(ucomp) * nNeurons[type]);
		cudaMalloc ((void **) &(synData->nGeneratedSpikesDevice[type]), sizeof(ucomp) * nNeurons[type]);

		for (int neuron = 0; neuron < nNeurons[type]; neuron++ ) {
			HinesStruct & h = hList[type][neuron];
			h.spikeTimes  = synData->spikeTimeListDevice[type] + spikeTimeListSize * neuron;
			h.nGeneratedSpikes = synData->nGeneratedSpikesDevice[type];// + neuron;
		}
	}

}

int prepareExecution(HinesMatrix *matrixList, int nNeurons, int kernelSteps, HinesStruct **hListPtr, int type) {

	HinesStruct *hList = (HinesStruct *)malloc(nNeurons*sizeof(HinesStruct)); //new HinesStruct[nNeurons];

	HinesMatrix & m0 = matrixList[0];
	int nComp = m0.nComp;
	int nCompActive = m0.activeChannels->getCompListSize();
	int nSynaptic = m0.synapticChannels->synapseListSize;

	int sharedMemMatrix    = sizeof(ftype) * (3*nComp + m0.mulListSize + m0.leftListSize); //+ nComp*nComp;
	int sharedMemSynaptic  = sizeof(ftype) * 4 * m0.synapticChannels->nChannelTypes;
	int sharedMemTotal = sharedMemMatrix + sharedMemSynaptic;

	int exclusiveMemMatrix   = sizeof(ftype) * (5*nComp + m0.leftListSize);
	int exclusiveMemActive   = sizeof(ftype) * (nCompActive*7);
	//int exclusiveMemSynaptic = sizeof(ftype) * m.spikeTimeListSize;
	int exclusiveMemTotal = exclusiveMemMatrix + exclusiveMemActive + sizeof(ftype)*nComp*kernelSteps;

	/**
	 * Allocates memory to all neurons
	 */
	printf("TotalMem = %10.3f MB\n",(sharedMemTotal + exclusiveMemTotal * nNeurons)/(1024.*1024.));
	ftype *memory;
	cudaMalloc((void **)(&(memory)), sharedMemTotal + exclusiveMemTotal * nNeurons);

	/**
	 * Copies the shared ftype data from matrix[0] to the GPU
	 */
	ftype *sharedMemMatrixAddress = m0.Cm;
	cudaMemcpy(memory, sharedMemMatrixAddress, sharedMemMatrix, cudaMemcpyHostToDevice);
	ftype *sharedMemSynapticAddress = m0.synapticChannels->tau;
	cudaMemcpy(memory+sharedMemMatrix/sizeof(ftype), sharedMemSynapticAddress, sharedMemSynaptic, cudaMemcpyHostToDevice);

	/**
	 * Allocates and copies shared data of type ucomp from matrix[0] to the GPU
	 */
	ucomp *ucompSharedMemory;
	cudaMalloc((void **)(&(ucompSharedMemory)), sizeof(ucomp) * ((m0.mulListSize + m0.leftListSize) * 2 + nComp + nCompActive + 3*nSynaptic));
	ucomp *ucompActiveAddress   = ucompSharedMemory + (m0.mulListSize + m0.leftListSize) * 2 + nComp;
	ucomp *ucompSynapticAddress = ucompActiveAddress + nCompActive;

	cudaMemcpy(ucompSharedMemory,    m0.ucompMemory, sizeof(ucomp) * ((m0.mulListSize + m0.leftListSize) * 2 + nComp), cudaMemcpyHostToDevice);
	cudaMemcpy(ucompActiveAddress,   m0.activeChannels->getCompList(), sizeof(ucomp) * nCompActive, cudaMemcpyHostToDevice);
	cudaMemcpy(ucompSynapticAddress, m0.synapticChannels->synapseCompList, sizeof(ucomp) * nSynaptic*3, cudaMemcpyHostToDevice);

	/**
	 * Prepare the MatrixStruct h for each neuron in the GPU
	 */
	for (int neuron = 0; neuron < nNeurons; neuron++ ) {

		HinesMatrix & m = matrixList[neuron];

		HinesStruct & h = hList[neuron];
		h.type = type;

		/**
		 * Shared memory
		 */
		h.memoryS = memory;
		h.Cm = h.memoryS;
		h.Ra = h.Cm + nComp;
		h.Rm = h.Ra + nComp;
		h.leftList = h.Rm + nComp;
		h.mulList  = h.leftList + m.leftListSize; // Used only when triangAll = 0

		/**
		 * Memory allocated per thread
		 */
		h.memoryE = memory + sharedMemTotal/sizeof(ftype) + neuron*exclusiveMemTotal/sizeof(ftype);
		ftype *exclusiveAddressM = m.rhsM;
		cudaMemcpy(h.memoryE, exclusiveAddressM, exclusiveMemMatrix, cudaMemcpyHostToDevice);
		// must match the order in HinesMatrix.cpp
		h.rhsM = h.memoryE	;
		h.vmList = h.rhsM + nComp;
		h.vmTmp = h.vmList + nComp;
		h.curr = h.vmTmp + nComp;
		h.active = h.curr + nComp;
		h.triangList = h.active + nComp; // triangularized list
		h.vmTimeSerie = h.triangList + m.leftListSize;

		h.currStep = m.currStep;
		h.vRest = m.vRest;
		h.dx = m.dx;
		h.nComp = m.nComp;
		h.dt = m.dt;
		h.triangAll = m.triangAll;

		h.mulListSize = m.mulListSize;
		h.leftListSize = m.leftListSize;
		h.mulListComp    = ucompSharedMemory;
		h.mulListDest    = h.mulListComp  + h.mulListSize;
		h.leftListLine   = h.mulListDest  + h.mulListSize;
		h.leftListColumn = h.leftListLine + h.leftListSize;
		h.leftStartPos   = h.leftListColumn + h.leftListSize;

		// must match the order in ActiveChannels.cpp
		if (nCompActive > 0) {
			cudaMemcpy(h.vmTimeSerie + nComp*kernelSteps, m.activeChannels->memory, exclusiveMemActive, cudaMemcpyHostToDevice);

			h.n = h.vmTimeSerie + nComp*kernelSteps;
			h.h = h.n + nCompActive;
			h.m = h.h + nCompActive;
			h.gNaBar = h.m + nCompActive;
			h.gKBar  = h.gNaBar + nCompActive;
			h.gNaChannel  = h.gKBar + nCompActive;
			h.gKChannel  = h.gNaChannel + nCompActive;

			h.ELeak = m.activeChannels->ELeak;
			h.EK = m.activeChannels->EK;
			h.ENa = m.activeChannels->ENa;

			h.compListSize = nCompActive;
			h.compList = ucompActiveAddress;

			checkCUDAError("Memory Allocation3:");
		}

		if (m.synapticChannels != 0) {

			h.synapseListSize = m.synapticChannels->synapseListSize;
			h.nChannelTypes   = m.synapticChannels->nChannelTypes;
			h.synapseCompList = ucompSynapticAddress;
			h.synapseTypeList = h.synapseCompList + h.synapseListSize;
			//h.synSpikeListPos = h.synapseTypeList + h.synapseListSize;

			h.spikeListSize 	= 0;
			h.spikeList     	= 0;
			h.synapseWeightList = 0;

			h.tau = memory+sharedMemMatrix/sizeof(ftype);
			h.gmax = h.tau  + 2 * m.synapticChannels->nChannelTypes;
			h.esyn = h.gmax + m.synapticChannels->nChannelTypes;

			h.lastSpike 		= m.lastSpike;
			h.spikeTimeListSize = m.spikeTimeListSize;
				h.threshold        = m.threshold;
				h.minSpikeInterval = m.minSpikeInterval;

		}
		matrixList[neuron].freeMem();
	}

	*hListPtr = hList;

	return 0;
}

/***************************************************************************
 * This part is executed in every integration step
 ***************************************************************************/
int launchGpuExecution(SharedNeuronGpuData *sharedData, int *nNeurons, int startType, int endType, int totalTypes, int threadNumber) {

	HinesMatrix **matrixList = sharedData->matrixList;

	int nDevices = 0;
	cudaGetDeviceCount(&nDevices);
	cudaSetDevice((threadNumber+1)%nDevices);
	checkCUDAError("Device selection:");

	ftype totalTime = sharedData->totalTime;
	int kernelSteps = 100;

	FILE *nSpkfile 	  = fopen("nSpikeKernel.dat", "w");
	FILE *lastSpkfile = fopen("lastSpikeKernel.dat", "w");

	/**
	 * Only one Gpu should update it
	 */
	pthread_mutex_lock (sharedData->mutex);
	if (sharedData->hList == 0) {
		sharedData->hList   = new HinesStruct *[totalTypes];
		sharedData->hGpu    = new HinesStruct *[totalTypes];
		sharedData->synData = new SynapticData;
		sharedData->synData->spikeListDevice = 0;
		sharedData->synData->totalTypes = totalTypes;

		if (threadNumber == 0) {
			bench.totalHinesKernel = 0;
			bench.totalConnRead    = 0;
			bench.totalConnWait    = 0;
			bench.totalConnWrite   = 0;
		}
	}
	pthread_mutex_unlock (sharedData->mutex);

	HinesStruct **hList   = sharedData->hList;
	HinesStruct **hGpu  = sharedData->hGpu;

	// Allocates the memory on the GPU and transfers the data
	for (int type = startType; type < endType; type++) {
		prepareExecution(matrixList[type], nNeurons[type], kernelSteps, &(hList[type]), type);
		if (benchConf.verbose)
			printf("GPU allocation with %d neurons, %d comparts on device %d.\n", nNeurons[type], matrixList[type][0].nComp, (threadNumber+1)%nDevices);
	}

	ftype dt = matrixList[startType][0].dt;
	int nSteps = totalTime / dt;

	prepareSynapses(sharedData, nNeurons, startType, endType);
	SynapticData *synData = sharedData->synData;

	/**
	 * Send data to Gpu
	 */
	for (int type = startType; type < endType; type++) {
		cudaMalloc((void **)&(hGpu[type]), sizeof(HinesStruct)*nNeurons[type]);
		cudaMemcpy(hGpu[type], hList[type], sizeof(HinesStruct)*nNeurons[type], cudaMemcpyHostToDevice);
		checkCUDAError("Memory Allocation:");
	}

	int nComp0 = matrixList[startType][0].nComp;
	int timeSerieMemSize = sizeof(ftype) * (nComp0 * kernelSteps);

	/**
	 * Perform the execution
	 */
	int nThreads = 0;
	if (nComp0 <= 4) nThreads = (sizeof (ftype) == 4) ? 196 : 96;
	else if (nComp0 <= 8) nThreads = (sizeof (ftype) == 4) ? 128 : 64;
	else if (nComp0 <= 12) nThreads = (sizeof (ftype) == 4) ? 96 : 32;
	else if (nComp0 <= 16) nThreads = (sizeof (ftype) == 4) ? 64 : 32;

	nThreads = 64;

	int nBlocks = nNeurons[startType]/nThreads;
	if (nNeurons[startType]%nThreads != 0) nBlocks++;
	if (nBlocks == 1) nThreads = nNeurons[startType];
	printf("Launching GPU kernel with %d blocks and %d threads per block types:%d-%d on device %d.\n", nBlocks, nThreads, startType, endType, (threadNumber+1)%nDevices);

	FILE *outFile;
	outFile = fopen("sampleVm.dat", "w");

	FILE *vmKernelFile;
	vmKernelFile = fopen("vmKernel.dat", "w");

	int nVmTimeSeries = 4;
	ftype **vmTimeSerie  = (ftype **)malloc(sizeof(ftype *) * nVmTimeSeries);
	for (int k=0; k<nVmTimeSeries; k++)
		vmTimeSerie[k] = (ftype *)malloc(timeSerieMemSize);

	int sharedMemorySize = 15 * 1024; // Valid for CUDA 1.3 (16kb)

	if (threadNumber == 0) {
		sharedData->connection = new Connections();
		//connection->createTestConnections(); # 100./nPyramidal, 100./nInhibitory,
		sharedData->connection->connectRandom (sharedData->pyrConnRatio, sharedData->inhConnRatio,
				sharedData->typeList, totalTypes, nNeurons, sharedData, threadNumber);
	}

	/**
	 * Synchronize threads before beginning
	 */
	pthread_mutex_lock (sharedData->mutex);
	sharedData->nBarrier++;
	if (sharedData->nBarrier < sharedData->nThreadsCpu)
		pthread_cond_wait(sharedData->cond, sharedData->mutex);
	else {
		sharedData->nBarrier = 0;
		pthread_cond_broadcast(sharedData->cond);
	}
	pthread_mutex_unlock (sharedData->mutex);

	if (threadNumber == 0) {
		bench.execPrepare  = gettimeInMilli();
		bench.execPrepareF = (bench.execPrepare - bench.matrixSetup)/1000.;
	}

	/**
	 * Solves the matrix for n steps
	 */
	for (int kStep = 0; kStep < nSteps; kStep += kernelSteps) {

		//printf ("F %d %d \n", startType, threadNumber);

		if (threadNumber == 0)
			bench.kernelStart  = gettimeInMilli();

		if (threadNumber == 0)
			printf("Starting Kernel %d -----------> %d \n", threadNumber, kStep);

		for (int type = startType; type < endType; type++) {
			checkCUDAError("Before SolveMatrixG Kernel:");
			solveMatrixG<<<nBlocks,nThreads, sharedMemorySize>>>(hGpu[type], kernelSteps, nNeurons[type],
					synData->spikeListDevice[type], synData->weightListDevice[type],
					synData->spikeListPosDevice[type], synData->spikeListStartDevice[type],
					synData->vmListDevice[type]);
			checkCUDAError("SolveMatrixG Kernel:");
		}

		checkCUDAError("Before Synapses:");

		if (benchConf.verbose == 1) {
			cudaThreadSynchronize(); // TODO: Retirar para fazer experimentos RYCEXP
			checkCUDAError("Sync:");
			printf("Finished Kernel %d -----------> %d \n", threadNumber, kStep);
		}

		/**
		 * Check if Vm is ok for all neurons
		 */
		if (benchConf.assertResultsAll == 1 || benchConf.printAllVmKernelFinish == 1) {
			for (int type = startType; type < endType; type++) {

				cudaMemcpy(synData->vmListHost[type], synData->vmListDevice[type], sizeof(ftype) * nNeurons[type], cudaMemcpyDeviceToHost);

				if (benchConf.assertResultsAll == 1) {
					for (int n = 0; n < nNeurons[type]; n++) {
						if ( synData->vmListHost[type][n] < -500 || 500 < synData->vmListHost[type][n] || synData->vmListHost[type][n] == 0.000000000000) {
							printf("********* type=%d neuron=%d %.2f\n", type, n, synData->vmListHost[type][n]);
							assert(false);
						}
					}
				}
			}
		}

		if (benchConf.verbose == 1)
			printf("Getting spikes %d\n", threadNumber);

		if (threadNumber == 0)
			bench.kernelFinish = gettimeInMilli();

		//printf("StartComm %d\n", threadNumber);

		/**
		 * Reads information from spike sources
		 * */
		for (int type = startType; type < endType; type++) {

			HinesStruct & hType = hList[type][0];

			checkCUDAError("Synapses0:");
			cudaMemcpy(synData->spikeTimeListHost[type],    synData->spikeTimeListDevice[type],    sizeof(ftype) * hType.spikeTimeListSize * nNeurons[type], cudaMemcpyDeviceToHost);
			checkCUDAError("Synapses1:");
			cudaMemcpy(synData->nGeneratedSpikesHost[type], synData->nGeneratedSpikesDevice[type], sizeof(ucomp) * nNeurons[type], cudaMemcpyDeviceToHost);
			checkCUDAError("Synapses2:");

			cudaThreadSynchronize(); // Used to prevent reads of local memory before the copy

			//printf("BeforeDispatch %d\n", threadNumber);

			for (int source = 0; source < nNeurons[type]; source++) {

				ucomp nGeneratedSpikes = synData->nGeneratedSpikesHost[type][source];
				//printf ("a -> type=%d source=%d nSpikes=%d \n", type, source, nGeneratedSpikes);
				if (nGeneratedSpikes > 0) {

					ftype *spikeTimes = synData->spikeTimeListHost[type] + hType.spikeTimeListSize * source;

					// Used to print spike statistics in the end of the simulation
					//if (sharedData->typeList[type] == PYRAMIDAL_CELL)
					sharedData->spkStat->addGeneratedSpikes(type, source, spikeTimes, nGeneratedSpikes);

					std::vector<Conn *> & connList = sharedData->connection->getConnArray(source + type*CONN_NEURON_TYPE);
					for (int conn=0; conn<connList.size(); conn++) {
						Conn *connStruct = connList[conn];
						SynapticChannels *targetSynapse = matrixList[ connStruct->dest / CONN_NEURON_TYPE ][ connStruct->dest % CONN_NEURON_TYPE ].synapticChannels;
						targetSynapse->addSpikeList(connStruct->synapse, nGeneratedSpikes, spikeTimes, connStruct->delay, connStruct->weigth);
					}

				}
			}
		}

		//printf("BeforeFree %d\n", threadNumber);

		for (int type = startType; type < endType; type++) {
			if (kStep > 0) cudaFree(synData->spikeListDevice[type]);
			if (kStep > 0) free(synData->spikeListGlobal[type]);
			checkCUDAError("cp0:");
		}

		if (threadNumber == 0)
			bench.connRead = gettimeInMilli();

		if (benchConf.verbose == 1)
			printf("Synchronizing %d\n", threadNumber);

		//printf("BeforeSync %d\n", threadNumber);

		/**
		 * Synchronize threads before communication
		 */
		pthread_mutex_lock (sharedData->mutex);
		sharedData->nBarrier++;
		if (sharedData->nBarrier < sharedData->nThreadsCpu) {
			pthread_cond_wait(sharedData->cond, sharedData->mutex);
		}
		else {
			sharedData->nBarrier = 0;
			pthread_cond_broadcast(sharedData->cond);
		}
		pthread_mutex_unlock (sharedData->mutex);

		if (threadNumber == 0)
			bench.connWait = gettimeInMilli();

		/**
		 * Prints the Vm at the end of each kernel execution
		 */
		if (threadNumber == 0 && benchConf.printAllVmKernelFinish == 1) {
			for (int type = 0; type < totalTypes; type++) {
				fprintf(vmKernelFile, "dt=%-10.2f\ttype=%d\t", dt * (kStep + kernelSteps), type);
				for (int n = 0; n < nNeurons[type]; n++)
					fprintf(vmKernelFile, "%10.2f\t", sharedData->synData->vmListHost[type][n]);
				fprintf(vmKernelFile, "\n");
			}
		}


		//printf ("J %d %d \n", startType, threadNumber);

		if (benchConf.verbose == 1)
			printf("Updating SpikeList %d\n", threadNumber);

		/**
		 * Updates the information in the target neurons
		 */
		for (int type = startType; type < endType; type++) {

			int totalNumberSpikes = 0;
			for (int n = 0; n < nNeurons[type]; n++) {
				HinesMatrix & m = matrixList[type][n];

				// Add some random spikes
				int nRandom = 0;
				if (sharedData->inputSpikeRate > 0 && sharedData->typeList[type] == PYRAMIDAL_CELL) {
					int32_t spkTime;
					for (int t = 0; t < 10; t++) { // At each ms
						random_r(sharedData->randBuf[threadNumber], &spkTime);
						if (spkTime / (float) RAND_MAX < sharedData->inputSpikeRate) {
							nRandom++;
							m.synapticChannels->addSpike(
									0,(kStep + kernelSteps)*dt + t + (10*(float)spkTime/RAND_MAX), 1);
						}
					}
				}
				// Fires only at some predefines times
				else if (sharedData->inputSpikeRate < 0 && sharedData->typeList[type] == PYRAMIDAL_CELL) {
					if (n%3 == 1) {
						m.synapticChannels->addSpike(0,(kStep + kernelSteps)*dt, 1);
					}
					if (n%3 == 2) {
						m.synapticChannels->addSpike(0,(kStep + kernelSteps)*dt, 1);
						m.synapticChannels->addSpike(0,(kStep + 1.5*kernelSteps)*dt, 1);
					}
				}

				// updates the spike list
				//if (type == 0 && n == 0) m.synapticChannels->printSpikeTimes(dt*(kStep+kernelSteps));
				m.synapticChannels->updateSpikeList(dt*(kStep+kernelSteps));
				//if (type == 0 && n == 0) m.synapticChannels->printSpikeTimes(dt*(kStep+kernelSteps));
				totalNumberSpikes += m.synapticChannels->spikeListSize;

				if (type == 1 && n == 1 && benchConf.verbose == 1)
					printf ("spikeListSize=%d nRandom=%d\n",m.synapticChannels->spikeListSize, nRandom );


				// Used to print spike statistics in the end of the simulation
				//if (sharedData->typeList[type] == PYRAMIDAL_CELL)
				sharedData->spkStat->addReceivedSpikes(
						type, n, m.synapticChannels->getAndResetNumberOfAddedSpikes()-nRandom);

				//if (type==0 && n == 1)
				//	m.synapticChannels->printSpikes(0);
			}

			if (benchConf.verbose == 1)
				printf("---> Allocating %f MB in GPU for %d spikes at %f for type %d.\n",
					sizeof(ftype) * totalNumberSpikes * 2/ (1024.*1024.), totalNumberSpikes, dt*(kStep+kernelSteps), type);

			/**
			 * Prepare the spike list in the format used in the GPU
			 */
			synData->spikeListGlobal[type] = (ftype *) malloc(sizeof(ftype) * totalNumberSpikes * 2);
			cudaMalloc ((void **) &(synData->spikeListDevice[type]), sizeof(ftype) * totalNumberSpikes * 2);
			synData->weightListDevice[type] = synData->spikeListDevice[type] + totalNumberSpikes;

			checkCUDAError("cp1:");

			ftype *spikeListTmp  = synData->spikeListGlobal[type];
			ftype *weightListTmp = synData->spikeListGlobal[type] + totalNumberSpikes;

			int spikeListPosTmp = 0;
			int spikePosListSize = matrixList[type][0].synapticChannels->synapseListSize;

			for (int n = 0; n < nNeurons[type]; n++) {

				HinesMatrix & m = matrixList[type][n];
				int mSpikeListSize = m.synapticChannels->spikeListSize;

				//				if ((type == 3 && (n == 2291 || n == 2289 || n == 2290)) || (type == 1 && n == 980))
				//					printf ("========================================> [%d][%d] spikes=%d vm=%.2f\n",
				//							type, n, m.synapticChannels->spikeListSize, synData->vmListHost[type][n]);

				memcpy(spikeListTmp, m.synapticChannels->spikeList, sizeof(ftype) * mSpikeListSize);
				memcpy(weightListTmp, m.synapticChannels->synapseWeightList, sizeof(ftype) * mSpikeListSize);

				if (n==0) synData->spikeListStartGlobal[type][0] = 0;
				synData->spikeListStartGlobal[type][n+1] = mSpikeListSize + synData->spikeListStartGlobal[type][n];

				for (int i=0; i<spikePosListSize; i++)
					synData->spikeListPosGlobal[type][spikeListPosTmp + i] =
							m.synapticChannels->synSpikeListPos[i];// + (spikeListTmp - synData->spikeListGlobal[type]);//sizeof(ftype);
				spikeListPosTmp += spikePosListSize;

				spikeListTmp  += mSpikeListSize;
				weightListTmp += mSpikeListSize;
			}
			synData->spikeListPosGlobal[type][spikeListPosTmp] = spikeListTmp - synData->spikeListGlobal[type];

			//printf("\n");
			//if (type ==1) {
			//	for (int n = synData->spikeListStartGlobal[type][2]; n < synData->spikeListStartGlobal[type][3]; n++)
			//		printf ("%f ", synData->spikeListGlobal[type][n]);
			//	printf("\n");
			//}

			//			printf("n--> ");
			//			for (int i=0; i<(1 + spikePosListSize * nNeurons[type]); i++)
			//				printf ("%d ", synData->spikeListPosGlobal[type][i]);
			//			printf("\n");

			/**
			 * Copy synaptic and spike info into the GPU
			 */
			cudaMemcpy(synData->spikeListDevice[type], synData->spikeListGlobal[type],
					sizeof(ftype)*totalNumberSpikes*2, cudaMemcpyHostToDevice);
			cudaMemcpy(synData->spikeListPosDevice[type], synData->spikeListPosGlobal[type],
					sizeof(int) * (1 + spikePosListSize * nNeurons[type]), cudaMemcpyHostToDevice);
			cudaMemcpy(synData->spikeListStartDevice[type], synData->spikeListStartGlobal[type],
					sizeof(int) * (nNeurons[type]+1), cudaMemcpyHostToDevice);
			checkCUDAError("cp2:");

		}

		if (threadNumber == 0)
			bench.connWrite = gettimeInMilli();

		if (benchConf.printSampleVms == 1) {
			if (benchConf.verbose == 1)
				printf("Writing Sample Vms thread=%d\n", threadNumber);

			int t1 = 0, n1 = 0;
			if (startType <= t1 && t1 < endType)
				cudaMemcpy(vmTimeSerie[0],  hList[t1][n1].vmTimeSerie, timeSerieMemSize, cudaMemcpyDeviceToHost);

			t1 = 0; n1 = 1;//2291;
			if (startType <= t1 && t1 < endType)
				cudaMemcpy(vmTimeSerie[1],  hList[t1][n1].vmTimeSerie, timeSerieMemSize, cudaMemcpyDeviceToHost);

			t1 = 0; n1 = 2;//135;
			if (startType <= t1 && t1 < endType)
				cudaMemcpy(vmTimeSerie[2],  hList[t1][n1].vmTimeSerie, timeSerieMemSize, cudaMemcpyDeviceToHost);

			t1 = 0; n1 = 3;//1203;
			if (startType <= t1 && t1 < endType)
				cudaMemcpy(vmTimeSerie[3],  hList[t1][n1].vmTimeSerie, timeSerieMemSize, cudaMemcpyDeviceToHost);
			checkCUDAError("Results obtaining:");

			for (int i = kStep; threadNumber == 0 && i < kStep + kernelSteps; i++) {
				fprintf(outFile, "%10.2f\t%10.2f\t%10.2f\t%10.2f\t%10.2f\n", dt * (i+1),
						vmTimeSerie[0][(i-kStep)], vmTimeSerie[1][(i-kStep)], vmTimeSerie[2][(i-kStep)], vmTimeSerie[3][(i-kStep)]);
			}
		}

		if (benchConf.printAllSpikeTimes == 1) {
			if (threadNumber == 0) // Uses only data from SpikeStatistics::addGeneratedSpikes
				sharedData->spkStat->printKernelSpikeStatistics(nSpkfile, lastSpkfile, (kStep+kernelSteps)*dt);
		}

		if (threadNumber == 0) {
			bench.totalHinesKernel	+= (bench.kernelFinish 	- bench.kernelStart)/1000.;
			bench.totalConnRead	  	+= (bench.connRead 		- bench.kernelFinish)/1000.;
			bench.totalConnWait		+= (bench.connWait 		- bench.connRead)/1000.;
			bench.totalConnWrite	+= (bench.connWrite 	- bench.connWait)/1000.;
		}


	} // (int kStep = 0; kStep < nSteps; kStep += kernelSteps)

	if (threadNumber == 0) {
		bench.execExecution  = gettimeInMilli();
		bench.execExecutionF = (bench.execExecution - bench.execPrepare)/1000.;
	}

	if (threadNumber == 0) {
		printf("%10.2f\t%10.5f\t%10.5f\n", dt * nSteps, (vmTimeSerie[0])[nComp0*kernelSteps-1], (vmTimeSerie[0])[kernelSteps-1]);
		printf("%10.2f\t%10.5f\t%10.5f\n", dt * nSteps, (vmTimeSerie[1])[nComp0*kernelSteps-1], (vmTimeSerie[1])[kernelSteps-1]);
	}

	// Used to print spike statistics in the end of the simulation
	if (threadNumber == 0)
		sharedData->spkStat->printSpikeStatistics("spikeGpu.dat", totalTime, bench);

	//ftype diff = (vmTimeSerie[nComp0*kernelSteps-1] - vmTimeSerieL[nComp0*kernelSteps-1]);
	//if ( (-0.0001 < diff && diff < 0.0001) == false )
	//	printf ("ERRRRRRRRRRRRRRROOOOOOOOOOOOOOOORRRRRRRRRRRRR!!!!\n");

	for (int type = startType; type < endType; type++) {
		cudaFree(synData->spikeListDevice[type]);
		free(synData->spikeListGlobal[type]);
	}
	// TODO: Free CUDA Memory
	fclose(outFile);
	free (vmTimeSerie);
	//delete[] hList;
	//delete synData;

	printf("Finished GPU execution\n");

	return 0;
}
